(load "ds-fp.scm")
(load "test.scm")
(load "macros.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;; TESTS for deployment of distributed aspects ;;;;;;;;;;

;IMPORTANT
;Read dist-examples.scm for info on distribution

;startup servers
(start-server-init "localhost" "8080"
                   '(begin 
                      (define count 0)
                      (define inc (fun () (set! count (+ 1 count)) count))
                      (define reset (fun () (set! count 0)))
                      (export-as inc "inc")))
(start-server "localhost" "9090") 

;; aspect applies 4 times:
; call(eval)->exec(eval)->call(thunk)->exec(thunk) 
(test (run (on-default 
             `(begin 
                (reset)
                (define aspect (list (fun (jp) (not (new? jp))) ;; match execs/calls
                                     (export (fun (jp) (inc))))) ;; remote advice!
                (deploy as-dynamic aspect 
                        ,(on-port "9090" '1))
                count)))
      4)

;; aspect applies 3 times:
; [call(eval)]->exec(eval)->call(thunk)->exec(thunk) 
(test (run (on-default 
             `(begin
                (reset)
                (define aspect (list (fun (jp) (not (new? jp)))
                                     (fun (jp) ((lookup "localhost" "8080" "inc"))))) ;; local advice!
                (deploy (list t f (fun (jp) (eq? (jp-port jp) "9090"))) 
                        aspect 
                        ,(on-port "9090" '1))
                count)))
      3)


;; aspect applies once, only on the non-remote call
; [call(eval)]->[exec(eval)]->call(thunk)->[exec(thunk)]
(test (run (on-default 
             `(begin 
                (reset)
                (define aspect (list (fun (jp) (and (call? jp)
                                                    (not (remote-call? jp))))
                                     (export (fun (jp) (inc)))))
                (deploy as-dynamic aspect 
                        ,(on-port "9090" '1))
                count)))
      1)

;; aspect applies once, only on the remote call to localhost:9090
; call(eval)->[exec(eval)]->[call(thunk)]->[exec(thunk)]
(test (run (on-default 
             `(begin 
                (reset)
                (define aspect (list (fun (jp) (and (remote-call? jp)
                                                    (eq? (host-port (jp-target-host jp)) "9090")))
                                     (export (fun (jp) (inc)))))
                (deploy as-dynamic aspect 
                        ,(on-port "9090" '1))
                count)))
      1)

;; export a remote print service on 9090, 
;; embed an aspect in it (for each exec)
;; call it from 9090 and 8080.
;; aspect applies twice.
(test (run 
       `(begin
          ,(on-default '(reset))
          ,(on-port "9090"
                    '(let ((remote-print (fun (x) (sprintf x))))
                       (let ((inc (lookup "localhost" "8080" "inc")))
                         (let ((aspect (list (exec remote-print)
                                             (fun (jp) (inc) (sprintf "*match*~n")))))
                           (sprintf "-----test without cflow (match twice)----~n")
                           (deploy-on (list t t t) aspect remote-print)
                           (export-as remote-print "remote-print")
                           (remote-print "-From 9090~n")))))
          ,(on-default '(begin
                          ((lookup "localhost" "9090" "remote-print") "-From 8080~n")
                          count))))
      2)

;; same as before, but using cflow/host to match 
;; only if in cflow of call from 8080
;; aspect applies once.
(test (run 
       `(begin
          ,(on-default '(reset))
          ,(on-port "9090"
                    '(let ((remote-print (fun (x) (sprintf x))))
                       (let ((inc (lookup "localhost" "8080" "inc")))
                         (let ((aspect (list (&& (exec remote-print) (cflow (in-port "8080")))
                                             (fun (jp) (inc) (sprintf "*match*~n")))))
                           (sprintf "-----test with cflow (match once)----~n")
                           (deploy-on (list t t t) aspect remote-print)
                           (export-as remote-print "remote-print")
                           (remote-print "-From 9090~n")))))
          ,(on-default '(begin
                          ((lookup "localhost" "9090" "remote-print") "-From 8080~n")
                          count))))
      1)

;; just to look at all the generated join points (see consoles)
(test (run (on-default 
             `(begin
                (reset)
                (define aspect (list t (fun (jp) (sprintf "~a~n" (jp-string jp)))))
                (deploy as-dynamic aspect 
                        ,(on-port "9090" '(fun (x) 1)))
                count)))
      0)

;; aspect embedded in function
;; create a function, applies it locally, then applies it remotely 
;; and returns it from remote host and apply back locally.
;; use propagate-global and active-global (could be avoided, since this is a default)
;; matches three times: exec on local, then remote, then local
(test (run (on-default 
             `(begin 
                (reset)
                (define aspect (list t (export (fun (jp) (inc)))))
                (define foo (fun (a b) (+ a b)))
                (deploy-on (active-global (propagate-global ds-all)) aspect foo)
                (foo 2 3)
                (,(on-port "9090" '(begin (foo 4 5) foo)) 6 7)
                count)))
      3)

;; aspect embedded in function (same as above)
;; use propagate-local strategy
;; matches once: 1st exec on local (does not propagate if remote)
(test (run (on-default 
             `(begin 
                (reset)
                (define aspect (list t (export (fun (jp) (inc)))))
                (define foo (fun (a b) (+ a b)))
                (deploy-on (propagate-local ds-all)
                           aspect 
                           foo)
                (foo 2 3)
                (,(on-port "9090" '(begin (foo 4 5) foo)) 6 7)
                count)))
      1)

;; aspect embedded in function (same as above)
;; use active-local strategy
;; matches twice: all exec on local (does not apply if remote)
(test (run (on-default 
             `(begin 
                (reset)
                (define aspect (list t (export (fun (jp) (inc)))))
                (define foo (fun (a b) (+ a b)))
                (deploy-on (active-local ds-all) 
                           aspect 
                           foo)
                (foo 2 3)
                (,(on-port "9090" '(begin (foo 4 5) foo)) 6 7)
                count)))
      2)