;;; definition of the parser for the language
;; and primitives

(define (parse sexp)
  (cond ((or (number? sexp)
             (string? sexp)
             (eqv? sexp #t)
             (eqv? sexp #f)) (make-lit-exp sexp))
  
	  ((and (list? sexp) (eqv? (car sexp) 'quote))
	   (make-lit-exp (cadr sexp)))	       
	       
	  ((symbol? sexp) (make-var-exp sexp))
          
          ((list? sexp)
           (let ((prim (assq (car sexp) *primitives*)))
	     (if prim
		 (make-prim-exp (cadr prim) (parse-exps (cdr sexp)))
                 
                 (case (car sexp)
                   ((if)
                    (make-if-exp (parse (cadr sexp))
                                 (parse (caddr sexp))
                                 (parse (cadddr sexp))))
                   
                   ;; (and .. .. ..)
                   ((and)
                    (make-and-exp (parse-exps (cdr sexp))))
                   
                   ;; (or .. .. ..)
                   ((or)
                    (make-or-exp (parse-exps (cdr sexp))))
                   
                   ;;(begin .. .. ..)
                   ((begin)
                    (make-begin-exp (parse-exps (cdr sexp))))
                   
                   ;;(define x 3)
                   ((define)
                    (make-def-exp (cadr sexp)
                                  (parse (caddr sexp))))
                   
                 ;; (let ((x 4) (y 3)) ...)
                   ((let)
                    (make-let-exp (map car (cadr sexp))
                                  (parse-exps (map cadr (cadr sexp)))
                                  (parse-exps (cddr sexp))))
                   ;; (set! x ...)
                   ((set!)
                    (make-set-exp (cadr sexp) (parse (caddr sexp))))
                   
                   ;; (fun (x y z) ...)
                   ((fun)
                    (make-fun-exp (cadr sexp)
                                  (parse-exps (cddr sexp))))
                   
                   
                   ;; (deploy-on (c d f) (pc adv) fun)
                   ((deploy-on)
                    (make-depl-on-exp (parse (cadr sexp))
                                      (parse (caddr sexp))
                                      (parse (cadddr sexp))))
                   
                   ;; (deploy (c d f) (pc adv) body)
                   ((deploy)
                    (make-depl-exp (parse (cadr sexp))
                                   (parse (caddr sexp))
                                   (parse-exps (cdddr sexp))))
                   ;; (this-jp)
                   ((this-jp) (make-this-jp-exp)) 
                   
                   (else
                    ;; (f 1 2 3)
                    (make-app-exp (parse (car sexp))
                                  (parse-exps (cdr sexp))))
                   ))))))

(define parse-exps (lambda (exps) (map parse exps)))

;;; primitives
(define *primitives*
  `((+       (lambda args (apply + args)))
    (-       (lambda args (apply - args)))
    (*       (lambda args (apply * args)))
    (/       (lambda args (apply / args)))
    (=       (lambda args (apply = args)))
    (<       (lambda args (apply < args)))
    (<=      (lambda args (apply <= args)))
    (>       (lambda args (apply > args)))
    (>=      (lambda args (apply >= args)))
    (zero?   (lambda args (apply zero? args)))
    (not     (lambda args (apply not args)))
    (eq?     (lambda args (apply equal? args)))
    (car     (lambda args (apply car  args)))
    (cdr     (lambda args (apply cdr  args)))
    (cadr    (lambda args (apply cadr args)))
    (caddr   (lambda args (apply caddr args)))
    (cons    (lambda args (apply cons args)))
    (list    (lambda args args))
    (length  (lambda args (apply length args)))
    (null?   (lambda args (apply null? args)))
    (map     (lambda args (apply map args)))
    (foldl   (lambda args (apply foldl args)))
    (foldr   (lambda args (apply foldr args)))
    (assq    (lambda args (apply assq args)))

    ; removed for implementation simplicity
    (write   (lambda args (apply write args)))
    (newline (lambda args (apply newline args)))   
    (printf  (lambda args (apply printf args)))

    ;; for looking up remote functions (returns function stub)
    (lookup (lambda args (apply lookup-function-on-host args)))
    
    ;; for exporting functions (both return function stub)
    ; anonymous
    (export (lambda args (apply export-function args)))
    ; with name
    (export-as (lambda args (apply export-named-function args)))

    ;; printf of the server console
    (sprintf  (lambda args (apply server-printf args)))

    ;; typical host accessors
    (host-name (lambda args (apply host-name args)))
    (host-port (lambda args (apply host-port args)))
    
    ;; for defining pointcuts and advices
    (jp-fun    (lambda (jp) (jp-fun jp)))
    (jp-kind   (lambda (jp) (jp-kind jp)))
    (jp-parent (lambda (jp) (jp-parent jp)))
    (jp-args   (lambda (jp) (jp-args jp)))
    (jp-host   (lambda (jp) (jp-host jp)))
    (jp-target-host (lambda (jp) (jp-target-host jp)))
    (jp-depth  (lambda (jp) (jp-depth jp)))
    (jp-string (lambda (jp) (jp-string jp)))
    ))